package net.grelf.grip;

// This Java source file is part of the application GRIP, for processing and
// measuring digital photographs.
// GRIP is available free from www.grelf.net - please always quote this URL in
// connection with GRIP.
//
// Copyright (c) Graham T Relf, UK, 2006
//
// All software contains bugs, even if, we hope, they are only extremely subtle
// ones. The author has tested this software to the best of his abilities to
// ensure that it is generally fit for its purpose of processing and measuring
// digital photographs. The author makes absolutely no claims and gives no
// guarantees whatsoever as to the accuracy of any output from this software.
// Anyone making decisions based on any output from this software has a
// responsibility to first convince themselves of the appropriateness of so
// doing. They should test the software for themselves and understand how it
// works, sufficiently for their needs. Source code and explanatory
// documentation are available for that purpose.

/** Batch processor. */
public class BatchProcessor
{
	private static final int MAX_STAR_BOUNDARY_SIZE = 250;
  
	private GRIP app;
  private BatchAction action;
  private final BatchMetadata batchMetadata = new BatchMetadata ();
  
  public BatchProcessor (BatchAction action)
  {
		this.app = GRIP.getInstance ();
    this.action = action;
  } // BatchProcessor

	/** Ask the user to select the files for processing, then process them
    * in the background, on a worker thread. */
	public void batch (String fileNameModifier,
                     java.util.Map <String, Object> params)
	{
		java.io.File [] files =
            net.grelf.FileIO.selectInputFiles ("Images to process");

		if (null != files)
		{
			doOpInBackground (files, fileNameModifier, params, files.length);
		}
	} // batch

	/* 14/11/07: Brought outside batch () in order to be accessible also for
   * pass 2 of the astro processes: */
	private net.grelf.Timer timer;
	private int nFiles;
	private int wd, ht; // Will get the width and height of the final image processed
	private int progress; // For monitor only.
	private boolean doFullAuto = false;
	private boolean showBlobsOnEachImage = false;
  private String resultPath;

  /* For "astro": */
	private AstroProcessKind astroProcessKind = null;
	private AstroProcessMethod astroProcessMethod = AstroProcessMethod.WARP; // default
	private net.grelf.image.Image darkIm = null;
    // Becomes non-null only if the user identifies a dark frame image.
  private net.grelf.image.Image flatIm = null;
    // Becomes non-null only if the user identifies a flat field image.
  private boolean doNeutralise = false;
	private boolean doBackground = false;
  private boolean doGnomonic = false;
	private int nBrightest;
	private int starSegmenterDifference;
	private int starSegmenterRadius;
	private BlobMeas [][] lookupBlobMeases = null;
	private java.util.List <java.util.List<ConnectionList>>
                                           connectionsInAllImages = null;
	private java.util.List <String> pathsToTempFiles = null;
	private java.util.List <net.grelf.PointFloat> centres = null;
                                            // Only for SHIFT_BRIGHTEST method
	private java.util.List<MatchPair> matches = null;

	/* For "astro", "accum", "average", "trail": */
	private Accumulator accum = null;

  /* For "average": */
  private int srcBitDepth = 16;

  /* For "warpgrid": */
  private WarpingGrid warpingGrid = null;

	/** Perform batch processing on the given array of files. This is meant to be
    * called from a worker thread, eg as set up by doOpInBackground ().
		* @param fileNameModifier Append this string to the name of the file when it
    * is saved after processing, so the original is not overwritten.
		* @param params Some batch processes expect parameters. These will have been
    * obtained before calling this method, either from a configuration file or
    * by asking the user. For example, rotation needs to know an angle. */
	public void batch (java.io.File [] files,
                     String fileNameModifier,
	                   java.util.Map <String, Object> params)
	{
		timer = new net.grelf.Timer ("BatchProcessor.batch", null);
		nFiles = files.length;
		wd = ht = 0; // Will get the width and height of the final image processed
		progress = 0; // For the monitor

    determineResultPath (params);
    determineNBrightest (params);
    determineSegmenterParameters ();

    double detectorWidth = Double.parseDouble	(
            Config.getValue ("Detector.width.mm.value"));
    double detectorHeight = Double.parseDouble	(
            Config.getValue ("Detector.height.mm.value"));

		double scale;

		try
		{
			scale = Math.abs (Double.parseDouble	(
              Config.getValue ("BackgroundFactor.value")));
		}
		catch (NumberFormatException ex)
		{
			scale = 1.0;
		}

		int cellFactor;

		try
		{
			cellFactor = Integer.parseInt (
              Config.getValue ("BackgroundCellFactor.value"));
		}
		catch (NumberFormatException ex)
		{
			cellFactor = 256;
		}

		/* For "crop" and "saturate": */
		ImFrame imframe = null;

		/* For infoxml": */
		java.io.PrintWriter out = null;

		/* For "multiply": */
		net.grelf.image.Image mulim = null;

		/* For "saturate": */
		int saturation = 0; // Leave as is

		/* For "subtract": */
		net.grelf.image.Image subim = null;

		///////////////////////////// Initialisation before the loop:

		if (this.action.equals (BatchAction.ASTRO))
		{
			astroProcessKind = (AstroProcessKind) params.get ("processKind");
			astroProcessMethod = (AstroProcessMethod) params.get ("method");

			pathsToTempFiles = new java.util.ArrayList <> ();
			centres = new java.util.ArrayList <> ();
			connectionsInAllImages =
              new java.util.ArrayList<> (); // A list for each image
			lookupBlobMeases = new BlobMeas [nFiles][nBrightest];

			String darkPath = (String) params.get ("darkPath");

			if (null != darkPath && darkPath.length () > 0)
			{
				darkIm = loadAuxiliaryImage (darkPath);
			}

			String flatPath = (String) params.get ("flatPath");

			if (null != flatPath && flatPath.length () > 0)
			{
				flatIm = loadAuxiliaryImage (flatPath);
			}

			String s = (String) params.get ("doNeutralise");
			doNeutralise = (null != s) && (s.equals ("true"));

			s = (String) params.get ("doBackground");
			doBackground = (null != s) && (s.equals ("true"));

			s = (String) params.get ("doGnomonic");
			doGnomonic = (null != s) && (s.equals ("true"));

			s = (String) params.get ("doFullAuto");
			doFullAuto = (null != s) && (s.equals ("true"));

			s = (String) params.get ("showBlobsOnEachImage");
			showBlobsOnEachImage = (null != s) && (s.equals ("true"));
		}
		else if (this.action.equals (BatchAction.DIVIDE))
		{
			flatIm = loadAuxiliaryImage ((String) params.get ("divPath"));
		}
		else if (this.action.equals (BatchAction.INFOXML))
		{
			try
			{
				out = new java.io.PrintWriter (
                new java.io.File ((String) params.get ("outputPath")));
				out.println ("<?xml version=\"1.0\"?>");
				out.println ("<Images>");
			}
			catch (java.io.IOException ex)
			{
				net.grelf.Util.warning ("Sorry", "Could not create XML output file");
				this.action = BatchAction.ABORT;
			}
		}
		else if (this.action.equals (BatchAction.MULTIPLY))
		{
			mulim = loadAuxiliaryImage ((String) params.get ("mulPath"));
		}
		else if (this.action.equals (BatchAction.SUBTRACT))
		{
			subim = loadAuxiliaryImage ((String) params.get ("subPath"));
		}

		////////////////////// Main loop through the image sequence:

		boolean averageRaw = false;
		// Can only become true if this.action is "accum", "average" or "trail" and
    // first image is raw and user then confirms

		net.grelf.image.Threshold manualThreshold = null;
    // Until it is perhaps needed.

		for (int imNo = 0; imNo < nFiles; imNo++)
		{
/****
GRIP.about ();
****/
			java.io.File file = files [imNo];

			if (file.exists ())
			{
				if (null != monitor)
				{
					monitor.setProgress (progress++);
					String pass;

					if (this.action.equals (BatchAction.ASTRO))
					{
						pass = "Pass 1: ";
					}
					else
					{
						pass = "";
					}

					monitor.setNote (pass + file.getName () + " [" + (imNo + 1) +
                  " of " + nFiles + "]");

					if (monitor.isCanceled ())
					{
						monitor.close ();
						monitor = null;
						return;
					}
				}

				net.grelf.image.Image image;

				if ((0 == imNo) 
         && (this.action.equals (BatchAction.CROP)
          || this.action.equals (BatchAction.SATURATE)))
				{
					imframe = new ImFrame (file.getPath (), new DefaultMeasurementHandler ());
					image = imframe.getImage ();
				}
				else
				{
					image = net.grelf.image.ImageLoader.load (file);
				}

				if (null == image)
        {
          net.grelf.Util.logWarning ("Image " + imNo + " null");
        }
        else
				{
          try
          {
            if ((imNo == 0) && image.isRaw ())
            {
              switch (this.action)
              {
                case ACCUM:
                case TRAIL:
                  averageRaw = net.grelf.Util.confirm (
                       "Accumulating raw images",
                       "Accumulate the raw images, without first interpreting?");
                  break;

                case AVERAGE:
                  averageRaw = net.grelf.Util.confirm (
                       "Averaging raw images",
                       "Average the raw images, without first interpreting?");
                  break;
              }
            }

            if (null != darkIm || null != flatIm || doNeutralise
                    || doBackground || doGnomonic)
            {
              if (!net.grelf.Util.isRoomForTemporaryImages (
                      file, nFiles, image.getWidth (), image.getHeight (),
                      image.getNBands (), image.getBitsPerChannel ()))
              {
                this.action = BatchAction.ABORT;
              }
            }

            if (!this.action.equals (BatchAction.ABORT))
            {
              boolean modified = false;

              if (image.isRaw () && !averageRaw)
              {
                Raw.interpret (image);
                Raw.removeMargin (image);
                modified = true;
              }

              wd = image.getWidth ();
              ht = image.getHeight ();

              if (this.action.equals (BatchAction.BIT8))
              {
                image = image.convertToImage8 ();

                if (null == image)
                {
                  this.action = BatchAction.ABORT;
                }
              }
              else
              if (this.action.equals (BatchAction.BIT16))
              {
                image = image.convertToImage16 ();

                if (null == image)
                {
                  this.action = BatchAction.ABORT;
                }
              }
              else
              if (this.action.equals (BatchAction.BIT32))
              {
                image = image.convertToImage32 ();

                if (null == image)
                {
                  this.action = BatchAction.ABORT;
                }
              }
              else
              if (this.action.equals (BatchAction.BIT64))
              {
                image = image.convertToImage64 ();

                if (null == image)
                {
                  this.action = BatchAction.ABORT;
                }
              }
              else
              if (this.action.equals (BatchAction.ASTRO))
              {
                net.grelf.image.Metadata metadata = batchMetadata.add (file, image);

                if (null != darkIm)
                {
                  try
                  {
                    image.subtractToZero (darkIm);
                    modified = true;
                  }
                  catch (net.grelf.image.IncompatibleImageException ex)
                  {
                    net.grelf.Util.warning ("Error", "Incompatible dark frame");
                    this.action = BatchAction.ABORT;
                  }
                }

                if (null != flatIm)
                {
                  image.divideByFlatField (flatIm);
                  modified = true;
                }

                if (doNeutralise)
                {
                  image.neutraliseBackground ();
                  modified = true;
                }

                if (doBackground)
                {
                  image.correctBackground (scale, cellFactor);
                  modified = true;
                }

                if (doGnomonic)
                {
                  int focalLength = determineFocalLength (metadata, params);

                  if (-1 != focalLength)
                  {
                    image = image.inverseGnomonicProjection (
                            focalLength, detectorWidth, detectorHeight);
                    modified = true;
                  }
                }

                if (modified)
                {
                  pathsToTempFiles.add (
                          saveTemporaryImage (file, fileNameModifier, image));
                }

                detectAndMeasureStars (manualThreshold, image, imNo, file);

                System.gc ();
              }
              else if (this.action.equals (BatchAction.AUTOSTRETCH))
              {
                image.autoStretch ((Boolean) params.get ("allSame"));
              }
              else if (this.action.equals (BatchAction.ACCUM)
                    || this.action.equals (BatchAction.AVERAGE)
                    || this.action.equals (BatchAction.TRAIL))
              {
                try
                {
                  astroAccumulate (imNo, image);
                }
                catch (net.grelf.image.IncompatibleImageException ex)
                {
                  net.grelf.Util.warning (app, "Error", "{0} - {1}",
                                          file.getPath (), ex.toString ());
                  this.action = BatchAction.ABORT;
                }
              }
              else if (this.action.equals (BatchAction.BACKGROUND))
              {
                image.correctBackground (scale, cellFactor);
              }
              else if (this.action.equals (BatchAction.CROP) && (0 == imNo))
              {
                BatchProcessor.cropFiles = files;
                imframe.getGP ().setDrawingMode (DrawingMode.CROP);
                imframe.setTitle ("Click to set rectangle corners");
                // The end of the cropping process is in
                // BatchProcessor.cropOthers().

                break; // No more files to process
              }
              else if (this.action.equals (BatchAction.DIVIDE))
              {
                image.divideByFlatField (flatIm);
              }
              else if (this.action.equals (BatchAction.FIT))
              {
                int width = (Integer) params.get ("width");
                int height = (Integer) params.get ("height");
                image.fit (width, height);
                fileNameModifier = "_" + image.getWidth () +
                                   "x" + image.getHeight ();
              }
              else if (this.action.equals (BatchAction.FLIPH))
              {
                image.flipHorizontal ();
              }
              else if (this.action.equals (BatchAction.FLIPV))
              {
                image.flipVertical ();
              }
              else if (this.action.equals (BatchAction.GNOMONIC))
              {
                image = image.gnomonicProjection (
                        image.getMetadata ().getFocalLength (),
                        detectorWidth, detectorHeight);
              }
              else if (this.action.equals (BatchAction.INFOXML))
              {
                infoXML (file, wd, ht, image, out);
              }
              else if (this.action.equals (BatchAction.INVERT))
              {
                image.invert ();
              }
              else if (this.action.equals (BatchAction.INVGNOMONIC))
              {
                image = image.inverseGnomonicProjection (
                        image.getMetadata ().getFocalLength (),
                        detectorWidth, detectorHeight);
              }
              else if (this.action.equals (BatchAction.MAGNIFY))
              {
                int factor = (Integer) params.get ("factor");
                image.scale (factor, false); // replicating
              }
              else if (this.action.equals (BatchAction.MONO))
              {
                image.convertToMonochrome ();
              }
              else if (this.action.equals (BatchAction.MULTIPLY))
              {
                image.multiply (mulim);
              }
              else if (this.action.equals (BatchAction.ROTATE))
              {
                image = rotate (params, image);
              }
              else if (this.action.equals (BatchAction.SATURATE))
              {
                if (0 == imNo)
                {
                  saturation = ImSaturateDialogue.askSaturate (imframe);

                  if (0 == saturation)
                  {
                    imNo = nFiles + 1;
                    // To leave the file loop and abort the batch process
                  }
                }
                else
                {
                  image.saturate (saturation);
                }
              }
              else if (this.action.equals (BatchAction.SCALE))
              {
                double factor = (Double) params.get ("factor");
                image.scale (factor, true); // Interpolating
              }
              else if (this.action.equals (BatchAction.SQUARE))
              {
                image.multiply (image);
              }
              else if (this.action.equals (BatchAction.SUBTRACT))
              {
                try
                {
                  image.subtract (subim);
                }
                catch (net.grelf.image.IncompatibleImageException ex)
                {
                  net.grelf.Util.warning ("Error",
                          "Incompatible image to subtract");
                  this.action = BatchAction.ABORT;
                }
              }
              else
              if (this.action.equals (BatchAction.WARPGRID))
              {
                warpImageUsingGrid (image);
              }

              if (null != fileNameModifier
              && !this.action.equals (BatchAction.ABORT)
              && !this.action.equals (BatchAction.ACCUM)
              && !this.action.equals (BatchAction.ASTRO)
              && !this.action.equals (BatchAction.AVERAGE)
              && !this.action.equals (BatchAction.CROP)
              && !(this.action.equals (BatchAction.SATURATE) && saturation == 0)
              && !this.action.equals (BatchAction.TRAIL))
              {
                saveResultingImage (file, image, fileNameModifier);
              }

              if ((0 == imNo) && this.action.equals (BatchAction.SATURATE))
              {
                imframe.setVisible (false);
                imframe.dispose ();
              }
            }

            image.dispose ();
            image = null; // Can clear away
          }
          catch (NullPointerException ex)
          {
            net.grelf.Util.logWarning (
                    "Image " + imNo + " failed:" + ex.toString ());
          }
				}
			}

			System.gc ();
		} // End of main loop through image files
    
    afterLoop (files, fileNameModifier, out);

		if (null != monitor)
		{
			monitor.close ();
			monitor = null;
		}

		timer.stop ();
	} // batch

  private void warpImageUsingGrid (net.grelf.image.Image image)
  {
    if (null == warpingGrid)
    {
      warpingGrid = new WarpingGrid ().load ();
      // Prompts user for file
    }

    if (null == warpingGrid)
    {
      net.grelf.Util.warning ("Sorry", "No warping grid is loaded");
      this.action = BatchAction.ABORT;
    }
    else
    {
      warpingGrid.use (image);
    }
  } // warpImageUsingGrid
  
  private void astroAccumulate (int imNo, net.grelf.image.Image image) 
  throws net.grelf.image.IncompatibleImageException
  {
    if (0 == imNo)
    {
      if (this.action.equals (BatchAction.TRAIL)
          && nFiles > 180) // Risk of overflow
      {
        accum = new Accumulator64 (image);
      }
      else
      {
        accum = new Accumulator32 (image);
      }
      
      srcBitDepth = image.getBitsPerChannel ();
    }
    else
    {
      if (this.action.equals (BatchAction.TRAIL))
      {
        accum.addWithStarBoost (image, nFiles);
      }
      else
      {
        accum.add (image);
      }
    }
  } // astroAccumulate

  private int determineFocalLength (
      net.grelf.image.Metadata metadata, java.util.Map <String, Object> params)
  {
    // Get metadata we need:
    int focalLength = Integer.MIN_VALUE;
    
    if (null != metadata)
    {
      focalLength = metadata.getFocalLength ();
    }
    
    if (focalLength == Integer.MIN_VALUE)
      // Meaning not found in the metadata
    {
      try
      {
        String s = (String) params.get ("focalLength");
        
        if (null == s || s.length () < 1)
        {
          focalLength = -1;
        }
        else
        {
          focalLength = Integer.parseInt (s);
          
          if (null != metadata)
          {
            metadata.setFocalLength (focalLength);
          }
        }
      }
      catch (NumberFormatException ex)
      {
        focalLength = -1;
      }
    }
    
    return focalLength;
  } // determineFocalLength
  
  private void determineSegmenterParameters ()
  {
    try
    {
      starSegmenterDifference = Integer.parseInt (
          Config.getValue ("StarSegmentDifference.value"));
      starSegmenterRadius = Integer.parseInt (
          Config.getValue ("StarSegmentRadius.value"));
    }
    catch (NumberFormatException ex)
    {
      String msg = "Config values missing for StarSegmenter parameters.\n" +
          "Try deleting file grip.properties and re-running GRIP.";
      net.grelf.Util.warning (null, "Error", msg);
    }
  } // determineSegmenterParameters


  private void determineNBrightest (java.util.Map <String, Object> params)
  {
    Object nb = params.get ("nBrightest");
    
    if (null != nb && nb instanceof Integer)
    {
      nBrightest = ((Integer) nb);
    }
    else
    {
      try
      {
        nBrightest = Integer.parseInt (Config.getValue ("NBrightest.value"));
      }
      catch (NumberFormatException ex)
      {
        nBrightest = 24;
      }
    }
  } // determineNBrightest

  private void determineResultPath (java.util.Map <String, Object> params)
  {
    Object resPath = params.get ("resultPath");
    
    if (null != resPath && resPath instanceof String)
    {
      this.resultPath = (String) resPath;
      
      if (!this.resultPath.endsWith (java.io.File.separator))
      {
        this.resultPath += java.io.File.separator;
      }
    }
    else
    {
      this.resultPath = "";
    }
  } // determineResultPath

  private void saveResultingImage (
      java.io.File file, net.grelf.image.Image image, String fileNameModifier)
  {
    String inputPath = file.getPath ();
    String extension = net.grelf.FileIO.getExtension (inputPath);
    
    // Cannot save raw:
    if (image.isRaw () || this.action.equals (BatchAction.RAW))
    {
      extension = ".tif";
    }
    
    int nBits = image.getBitsPerChannel ();
    
    if (nBits == 8)
    {
      extension = ".jpg";
    }
    else
    if (nBits > 16)
    {
      extension = ".fits";
    }
    
    String outputPath =
        net.grelf.FileIO.removeExtension (inputPath) +
        fileNameModifier + extension;
    image.save (outputPath);
  } // saveResultingImage

  private void detectAndMeasureStars (
      net.grelf.image.Threshold manualThreshold,
      net.grelf.image.Image image, int imNo, java.io.File file)
  {
    // Threshold and detect blobs:
    BlobMeasList meases;
    BlobMask mask;
    
    if (manualThreshold != null)
    {
      mask = (BlobMask) image.threshold (manualThreshold);
    }
    else
    {
      mask = StarSegmenter.segment (image,
          starSegmenterDifference, starSegmenterRadius);
    }
    
    if (null == mask)
    {
      net.grelf.Util.logInfo ("null segmented mask");
    }
    else
    {
      java.util.List <Blob> blobs = mask.detectBlobs (false);
                                            // Ignore 1-pixel blobs
      
      if (null == blobs)
      {
        net.grelf.Util.logInfo ("null blobs list");
      }
      else
      {
        /****/
        net.grelf.Util.logInfo ("NBLOBS = {0}", blobs.size ());
        /****/
        mask = null; // Can dispose
        
        if ((0 == blobs.size ())
            && (null == manualThreshold) && (0 == imNo))
        {
          net.grelf.Util.warning ("Error",
              "No objects were detected.\nTry thresholding this\nfirst image manually");
          ImFrame imf = new ImFrame ("First image",
              image.getWidth (), image.getHeight ());
          imf.getImPane ().setImage (image);
          manualThreshold = imf.threshold ();
          meases = imf.detectBlobs (false); // Ignore 1-pixel blobs
          imf.setVisible (false);
          imf.dispose ();
        }
        else
        {
          meases = new BlobMeasList (null); // It has no ImFrame
          // In descending order of brightness
          
          for (Blob blob : blobs)
          {
            java.util.List <java.awt.Point> boundary = blob.getBoundary ();
            
            if (null != boundary
                && boundary.size () < MAX_STAR_BOUNDARY_SIZE)
            {
              BlobMeas meas = blob.measure (image);
              
              if (null != meases)
              {
                meases.insert (meas);
              }
              
              ((Blob_) blob).nullRows ();
              // To save memory. Must come after measuring
            }
            else
            {
              blob = null; // Ensure disposal
            }
          }
          
          batchMetadata.addToNStars (blobs.size ());
          
          if (null != meases)
          {
            double sumCircs = 0.0;
            int nCircs = 5;
            
            // Measure circularity of brightest blobs:
            for (int i = 0; i < nCircs && i < meases.size (); i++)
            {
              BlobMeas meas = meases.get (i);
              sumCircs += meas.calculateCircularity ();
            }
            
            batchMetadata.addToMeanCircularities (
                (float) (sumCircs / nCircs));
          }
          
          blobs = null; // Can dispose
        }
        
        if (astroProcessMethod == AstroProcessMethod.SHIFT_BRIGHTEST)
        {
          if (meases.size () > 0)
          {
            BlobMeas brightestMeas = meases.get (0);
            brightestMeas.blob = null;
            // So the blob can be cleared away because it may be very big.
            
            centres.add (new net.grelf.PointFloat (
                brightestMeas.centre.x, brightestMeas.centre.y));
            // Clone the PointFloat so the BlobMeas can be cleared away.
          }
          
          meases = null;
        }
        else
        {
          // List the positions of the brightest:
          connectionsInAllImages.add (
              Analyser.getSortedConnections (meases, nBrightest));
          
          int blobNo = 0;
          
          for (BlobMeas meas : meases)
          {
            lookupBlobMeases [imNo][blobNo] = meas;
            blobNo++;
            
            if (blobNo >= nBrightest)
            {
              break;
            }
          }
          
          if (showBlobsOnEachImage)
          {
            displayBlobsInImage (file.getName (), meases);
          }
        }
        
        meases = null; // Ensure disposal
      }
    }
  } // detectAndMeasureStars

  /** Final processing, after looping through the images. */
  private void afterLoop (java.io.File[] files, 
                           String fileNameModifier, java.io.PrintWriter out)
  {
    if (this.action.equals (BatchAction.ASTRO))
    {
      if (astroProcessMethod == AstroProcessMethod.SHIFT_BRIGHTEST)
      {
        if (null != monitor)
        {
          monitor.close ();
          monitor = null;
        }
        
        if (0 == pathsToTempFiles.size ())
        {
          // Temporary files were not needed, so store the paths to the original
          // files for pass 2:
          for (int i = 0; i < nFiles; i++)
          {
            pathsToTempFiles.add (files [i].getPath ());
          }
        }
        
        astroPass2 ();
      }
      else // SHIFT_AVERAGE or WARP:
      {
        if (null != monitor)
        {
          monitor.setNote ("Analysing matches");
        }
        
        /* Diagnostic
        Connection.reportConnections (connectionsInAllImages, "connections.csv");
        */
        matches = Analyser.matchBrightest (connectionsInAllImages);
        
        if (null != monitor)
        {
          monitor.close ();
          monitor = null;
        }
        
        if (matches.isEmpty())
        {
          if (net.grelf.Util.confirm ("Sorry",
"No matches found between images.\nDo you want to see which\nobjects were detected?"))
          {
            displayAllBlobs (wd, ht, lookupBlobMeases);
            deleteTemporaryFiles (pathsToTempFiles);
          }
        }
        else
        {
          /* Diagnostic
          MatchPair.reportMatches (matches, lookupBlobMeases, nBrightest, "matches.csv");
          */
          int middleImNo = nFiles / 2;
          java.util.List<MatchList> listOfLists =
              MatchList.makeListOfLists (matches, nBrightest, middleImNo);
          /* Diagnostic
          MatchList.reportListOfLists (listOfLists, lookupBlobMeases, "lists.csv");
          */
          if (listOfLists.isEmpty ())
          {
            net.grelf.Util.message ("Abort",
                "No matching objects found across images");
            deleteTemporaryFiles (pathsToTempFiles);
          }
          else
          {
            if (0 == pathsToTempFiles.size ())
            {
              net.grelf.Util.logInfo ("Storing paths to " + nFiles + " original files");
              // Temporary files were not needed, so store the paths to
              // the original files for pass 2:
              for (int i = 0; i < nFiles; i++)
              {
                pathsToTempFiles.add (files [i].getPath ());
              }
            }
            
            if (null != monitor)
            {
              monitor.close ();
              monitor = null;
            }
            
            if (!doFullAuto)
            {
              MatchPair.drawMatches (listOfLists, matches, wd, ht,
                  lookupBlobMeases, this);
              // Will call astroPass2 () when user closes the window containing
              // the drawing
            }
            else
            {
              MatchPair.setListOfLists (listOfLists);
              astroPass2 ();
            }
          }
        }
      }
    }
    else
    if (null != fileNameModifier)
    {
      switch (this.action)
      {
        case AVERAGE:
          ImFrame result = new ImFrame (
              files [0].getPath (), new DefaultMeasurementHandler ());
          result.setCursor (GRIP.GRIP_WAIT_CURSOR);

          if (null != monitor)
          {
            monitor.setNote ("Dividing and drawing result");
          }

          accum.divide (nFiles);

          if (32 == srcBitDepth)
          {
            result.getImPane ().setImage (accum);
          }
          else
          {
            result.getImPane ().setImage (accum.convertToImage16 ());
          }

          String inputPath = files [0].getPath ();
          String outputPath =
              net.grelf.FileIO.removeExtension (inputPath) + fileNameModifier +
              net.grelf.FileIO.getExtension (inputPath);
          result.setCaption (new java.io.File (outputPath).getName ());
          result.redisplay ();
          result.getImage ().save (outputPath);
          result.setCursor (GRIP.GRIP_DEFAULT_CURSOR);
          GRIP.getTable ().refresh ();
          break;

        case ACCUM:
        case TRAIL:
          String firstFilePath = files [0].getPath ();
          String lastFileName = files [nFiles - 1].getName ();
          String outPath =
              net.grelf.FileIO.removeExtension (firstFilePath) + "-" +
              net.grelf.FileIO.removeExtension (lastFileName) + "_ave.fits";
          accum.save (outPath);
          new ImFrame (outPath, accum);
          break;
      }
    }
    else
    if (this.action.equals (BatchAction.INFOXML))
    {
      out.println ("</Images>");
      out.close ();
    }
  } // afterLoop

  private net.grelf.image.Image rotate (
      java.util.Map <String, Object> params, net.grelf.image.Image image)
  {
    double angle = (Double) params.get ("angle");
    
    if (89.99999 < angle && angle < 90.00001)
    {
      net.grelf.image.Image newIm = image.rotate90 ();
      image.dispose ();
      image = newIm;
    }
    else
    if (179.99999 < angle && angle < 180.00001)
    {
      net.grelf.image.Image newIm = image.rotate180 ();
      image.dispose ();
      image = newIm;
    }
    else
    if (269.99999 < angle && angle < 270.00001)
    {
      net.grelf.image.Image newIm = image.rotate270 ();
      image.dispose ();
      image = newIm;
    }
    else
    if (-89.99999 > angle && angle > -90.00001)
    {
      net.grelf.image.Image newIm = image.rotate270 ();
      image.dispose ();
      image = newIm;
    }
    else
    if (-179.99999 > angle && angle > -180.00001)
    {
      net.grelf.image.Image newIm = image.rotate180 ();
      image.dispose ();
      image = newIm;
    }
    else
    if (-269.99999 > angle && angle > -270.00001)
    {
      net.grelf.image.Image newIm = image.rotate90 ();
      image.dispose ();
      image = newIm;
    }
    else
    {
      image.rotate (angle);
    }

    return image;
  } // rotate

  private net.grelf.image.Image loadAuxiliaryImage (String path)
  {
    net.grelf.image.Image image =
            net.grelf.image.ImageLoader.load (new java.io.File (path));

    if (image.isRaw ())
    {
      Raw.interpret (image);
      Raw.removeMargin (image);
    }

    return image;
  } // loadAuxiliaryImage

	private void deleteTemporaryFiles (java.util.List<String> pathsToTempFiles)
	{
		// Delete any temporary files:
		for (String path : pathsToTempFiles)
		{
			if (path.endsWith ("_temp.fits"))
        // Avoids deleting the original input files
			{
				if (new java.io.File (path).delete ())
				{
					net.grelf.Util.logInfo ("Deleted {0}", path);
				}
				else
				{
					net.grelf.Util.logWarning ("FAILED to delete {0}", path);
				}
			}
		}
	} // deleteTemporaryFiles

  /** 12.9.16: Now saves in FITS format rather than TIFF. */
	private String saveTemporaryImage (java.io.File f, String fileNameModifier,
                                     net.grelf.image.Image image)
	{
		String pathToTempFile = 
            this.resultPath +
            net.grelf.FileIO.removeExtension (f.getName ()) +
            fileNameModifier + ".fits";
    image.saveAsFITS (pathToTempFile);
    net.grelf.Util.logInfo ("Saved {0}", pathToTempFile);
    return pathToTempFile;
	} // saveTemporaryImage

	/** Pass 2: for each adjusted image, "astro" action only */
	protected void astroPass2 ()
	{
		if (null != matches && matches.isEmpty ())
		{
			deleteTemporaryFiles (pathsToTempFiles);
		}
		else
		{
			// Dispose of any diagnostic blob detection frames:
			{
				boolean disposedSome = false;

				for (ImFrame imf : GRIP.getTable ().getList ())
				{
					if (-1 != imf.getCaption ().indexOf ("objects detected"))
					{
						imf.dispose ();
						disposedSome = true;
					}
				}

				if (disposedSome)
				{
					GRIP.getTable ().refresh ();
				}
			}

      boolean cancelled = false;
      boolean [] selected;
      int nIms = pathsToTempFiles.size ();

      if (doFullAuto || null == matches)
      // matches can be null for shift by brightest only
      {
        selected = new boolean [nIms];

        for (int i = 0; i < nIms; i++)
        {
          selected [i] = true;
        }
      }
      else // User may reject some images:
      {
        ImageSummaryDialogue de = new ImageSummaryDialogue (
                batchMetadata, matches, lookupBlobMeases, nBrightest, true,
                this.resultPath);

        selected = de.getSelected ();
        de.dispose ();

        if (null == selected)
        {
          cancelled = true;
        }
        else
        {
/****
for (int i = 0; i < selected.length; i++)
{
  System.out.println (selected [i]);
}
/****/
        }
      }

      if (!cancelled)
      {
  			progress = 0; // For the monitor
//        monitor = new javax.swing.ProgressMonitor (
//                null, "Batch [astro]", "Running", 0, nFiles);

        try
        {
          int nAccumulated = 0;
          int nFailed = 0;
          int nSelected = 0;

          for (int i = 0; i < selected.length; i++)
          {
            if (selected [i])
            {
              nSelected++;
            }
          }

          if (0 < nIms && 0 < nSelected)
          {
            timer = new net.grelf.Timer ("BatchProcessor.astroPass2", null);

            if (null != monitor)
            {
              monitor.setMaximum (nIms);
            }

            int middleImNo = nIms / 2;

            for (int imNo = 0; imNo < nIms; imNo++)
            {
              if (selected [imNo])
              {
                try
                {
                  String path = pathsToTempFiles.get (imNo);
                  java.io.File f = new java.io.File (path);

                  if (null != monitor)
                  {
                    monitor.setProgress (progress++);
                    monitor.setNote ("Pass 2: " + f.getName () + " [" +
                            (imNo + 1) + " of " + nIms + "]");

                    if (monitor.isCanceled ())
                    {
                      monitor.close ();
                      monitor = null;
                      return;
                    }
                  }

                  net.grelf.image.Image image =
                          loadAuxiliaryImage (f.getPath ());

                  if (null == image)
                  {
                    net.grelf.Util.warning ("Error",
                      "Could not reload image {0} from\n{1}",
                      imNo, f.getPath ());
                  }
                  else
                  {
                    if (null == accum
                     && astroProcessKind == AstroProcessKind.COMBINE)
                    {
                      accum = new Accumulator32 (
                              image.getWidth (), image.getHeight (),
                              image.getNBands ());
                      // 19/11/07: Don't use the other constructor because we
                      // don't need to set pixel values.
                      // We will be mapping all images onto the middle one, not
                      // the first.
                      nAccumulated = 0;
                    }

                    if (imNo == middleImNo
                     && astroProcessKind == AstroProcessKind.COMBINE)
                    {
                      accum.add (image);
                      nAccumulated++;
                    }
                    else
                    {
                      switch (astroProcessMethod)
                      {
                      case SHIFT_AVERAGE:
                        if (astroProcessKind == AstroProcessKind.COMBINE)
                        {
                          accum.averageShiftAccumulate (image, imNo, middleImNo,
                                  matches, lookupBlobMeases, nBrightest);
                          nAccumulated++;
                        }
                        else // AstroProcessKind.ALIGN:
                        {
                          net.grelf.image.Image shiftedImage = null;

                          if (imNo == middleImNo)
                          {
                            // No warping needed or possible:
                            shiftedImage = image;
                          }
                          else
                          {
                            double [] offset = MatchList.getAverageShift (
                                    imNo, middleImNo, matches,
                                    lookupBlobMeases, nBrightest);

                            if (null == offset || 2 > offset.length)
                            {
                              net.grelf.Util.warning ("Error",
                                "Could not find average shift for image " + imNo);
                            }
                            else
                            {
                              image.translate (offset [0], offset [1]);
                              shiftedImage = image;
                            }
                          }

                          if (null == shiftedImage)
                          {
                            net.grelf.Util.logWarning (
                              "Failed to shift image {0}", f.getName ());
                            nFailed++;
                          }
                          else
                          {
                            String inputName = f.getName ();

                            if (null == inputName)
                            {
                              net.grelf.Util.warning ("Error",
                                      "Null inputName");
                            }

                            String extension;

                            if (image.getBitsPerChannel () > 16)
                            {
                              extension = ".fits";
                            }
                            else
                            {
                              extension = ".tif";
                            }

                            String outputPath =
                                this.resultPath +
                                net.grelf.FileIO.removeExtension (inputName) +
                                "_shav" + extension;
                            shiftedImage.save (outputPath);
                            shiftedImage.dispose ();
                            shiftedImage = null;
                          }
                        }

                        break;

                      case SHIFT_BRIGHTEST:
                        if (astroProcessKind == AstroProcessKind.COMBINE)
                        {
                          if (accum.brightestShiftAccumulate (
                                  image, imNo, middleImNo, centres))
                          {
                            nAccumulated++;
                          }
                        }
                        else // AstroProcessKind.ALIGN:
                        {
                          // User cannot select this from AstroProcessDialogue
                        }

                        break;

                      default:
                        if (astroProcessKind == AstroProcessKind.COMBINE)
                        {
                          if (accum.warpAccumulate (image, imNo, middleImNo,
                                  matches, lookupBlobMeases, nBrightest))
                          {
                            nAccumulated++;
                          }
                        }
                        else // WARP each image and save it
                        {
                          net.grelf.image.Image warpedImage;

                          if (imNo == middleImNo)
                          {
                            // No warping needed or possible:
                            warpedImage = image;
                          }
                          else
                          {
                            warpedImage = Warp.warp (image, imNo, middleImNo,
                                   matches, lookupBlobMeases, nBrightest);
                          }

                          if (null == warpedImage)
                          {
                            net.grelf.Util.logWarning (
                                    "Failed to warp image {0}", f.getName ());
                            nFailed++;
                          }
                          else
                          {
                            String inputName = f.getName ();
                            String extension;

                            if (image.getBitsPerChannel () > 16)
                            {
                              extension = ".fits";
                            }
                            else
                            {
                              extension = ".tif";
                            }

                            String outputPath =
                                this.resultPath +
                                net.grelf.FileIO.removeExtension (inputName) + 
                                "_warp" + extension;
                            warpedImage.save (outputPath);
                            warpedImage.dispose ();
                            warpedImage = null;
                          }
                        }
                      }
                    }

                    image.dispose ();
                  }

                  image = null;
                  System.gc ();
                }
                catch (NullPointerException ex)
                {
                  net.grelf.Util.logWarning ("Image " + imNo +
                          " failed in pass 2: " + ex.toString ());
                }
              }
            }

            deleteTemporaryFiles (pathsToTempFiles);

            if (astroProcessKind == AstroProcessKind.COMBINE)
            {
              if (null != monitor)
              {
                monitor.close ();
                monitor = null;
              }

              net.grelf.image.Metadata metadata =
                      batchMetadata.setImageMetadata (accum, nAccumulated);
              String firstFileName = net.grelf.Util.removeSubString (
                  new java.io.File (pathsToTempFiles.get (0)).getName (),
                  "_temp");
              String lastFileName = net.grelf.Util.removeSubString (
                  new java.io.File (pathsToTempFiles.get (nIms - 1)).getName (),
                  "_temp");

              String method = "";

              switch (astroProcessMethod)
              {
                case SHIFT_AVERAGE:
                  method = "_shav";
                  metadata.addValue (net.grelf.image.Metadata.METHOD_KEYWORD,
                                     "Average shift");
                  break;

                case SHIFT_BRIGHTEST:
                  method = "_shbr";
                  metadata.addValue (net.grelf.image.Metadata.METHOD_KEYWORD,
                          "Brightest shift");
                  break;

                case WARP:
                  method = "_warp";
                  metadata.addValue (net.grelf.image.Metadata.METHOD_KEYWORD,
                          "Warp");
                  break;
              }

              String outRootPath = 
                  this.resultPath +
                  net.grelf.FileIO.removeExtension (firstFileName) + "-" +
                  net.grelf.FileIO.removeExtension (lastFileName) +
                      method + ".fits";
              ImFrame resultFrame = new ImFrame (outRootPath, accum);
              resultFrame.getImage ().save (outRootPath);
            }

            if (null != matches)
            {
              // Let user see drive errors etc, not seen at the end of pass 1,
              // but not modify the table:
              new ImageSummaryDialogue (batchMetadata, matches,
                      lookupBlobMeases, nBrightest, false, this.resultPath);
            }

            if (nFailed > 0)
            {
              net.grelf.Util.warning ("Sorry", nFailed +
                      " images could not be combined");
            }

            timer.stop ();
            net.grelf.Util.message ("Done", nSelected +
                    " images processed in pass 2");
          }
          else
          {
            net.grelf.Util.warning ("Sorry", "No files processed in pass 2");
          }
        }
        catch (net.grelf.image.IncompatibleImageException ex)
        {
          net.grelf.Util.warning (app, "Error", ex.toString ());
        }
      }
      else
      {
        net.grelf.Util.warning ("Sorry", "No files processed in pass 2");
      }

			MatchPair.resetStaticObjects (); // 29/9/09: Fix memory leak

			if (null != monitor)
			{
				monitor.close ();
				monitor = null;
			}

			pathsToTempFiles = null;
      batchMetadata.dispose ();
			centres = null;
		}
	} // astroPass2

/*
  private java.util.List <String> rejectImages (java.util.List <String> filePaths,
                                                boolean [] selected)
  {
    java.util.List <String> newFilePaths = new java.util.ArrayList <String> ();

    for (int i = 0; i < selected.length; i++)
    {
      if (selected [i])
      {
        newFilePaths.add (filePaths.get (i));
      }
    }

    filePaths.clear ();
    return newFilePaths;
  } // rejectImages
*/

	/* For crop: */
	private static java.io.File [] cropFiles = null;
  // Set above, when prompting user on first image
	//NB: making this static is a compromise to avoid having to pass either the
  // file list or an instance of BatchProcessor to more general interaction
  // methods. Users will not be able to run more than one
	// batch cropping process but they are unlikely to want to do so.
  public static boolean isBatchCropping () { return (null != cropFiles); }

	/** Nov 07: No longer static. */
	protected void cropOthers (java.awt.Point pt1, java.awt.Point pt2)
	{
    GRIP grip = GRIP.getInstance ();
    grip.setCursor (GRIP.GRIP_WAIT_CURSOR);
    grip.pack ();

    try
    {
      progress = 0; // For the monitor
      int nFilesToCrop = BatchProcessor.cropFiles.length;

      for (int imNo = 1; imNo < nFilesToCrop; imNo++)
      {
        java.io.File f = BatchProcessor.cropFiles [imNo];

        if (f.exists ())
        {
          if (null != monitor)
          {
            monitor.setProgress (imNo);
            monitor.setNote (f.getName () + " [" + (imNo + 1) +
                    " of " + nFilesToCrop + "]");

            if (monitor.isCanceled ())
            {
              monitor.close ();
              monitor = null;
              return;
            }
          }

          net.grelf.image.Image image = loadAuxiliaryImage (f.getPath ());

          if (null != image)
          {
            image.crop (pt1, pt2);
            String inputPath = f.getPath ();
            String outputPath =
                    net.grelf.FileIO.removeExtension (inputPath) + "_crop" +
                    net.grelf.FileIO.getExtension (inputPath);
            image.save (outputPath);
          }
        }
      }

      if (null != monitor)
      {
        monitor.close ();
        monitor = null;
      }
    }
    catch (Exception ex) // To ensure cropFiles is always nullified, below
    {
      net.grelf.Util.warning (grip, "Error", ex.toString ());
    }

		cropFiles = null;
		grip.setCursor (GRIP.GRIP_DEFAULT_CURSOR);
	} // cropOthers

	/** Output the information for a given image as XML elements. */
	private void infoXML (java.io.File f, int wd, int ht,
                        net.grelf.image.Image image, java.io.PrintWriter out)
	{
		out.println ("\t<Image>");
		out.println ("\t\t<Name>" + f.getName () + "</Name>");
		out.println ("\t\t<Path>" + f.getPath () + "</Path>");
		out.println ("\t\t<Width>" + wd + "</Width>");
		out.println ("\t\t<Height>" + ht + "</Height>");
		out.println ("\t\t<Channels>" + image.getNBands () + "</Channels>");
		out.println ("\t\t<BitsPerChannel>" + image.getBitsPerChannel () +
            "</BitsPerChannel>");
    net.grelf.image.Metadata metadata = image.getMetadata ();

    if (null != metadata)
    {
      java.util.Map<String,String> map = metadata.getMap ();

      if (map.containsKey ("Artist"))
      {
        out.println ("\t\t<Photographer>" + map.get ("Artist") +
                      "</Photographer>");
      }

      if (map.containsKey ("Make"))
      {
        out.println ("\t\t<CameraMake>" + map.get ("Make") + "</CameraMake>");
      }

      if (map.containsKey ("Model"))
      {
        out.println ("\t\t<CameraModel>" + map.get ("Model") + "</CameraModel>");
      }

      if (map.containsKey ("DateTimeOriginal"))
      {
        out.println ("\t\t<DateTimeOriginal>" + map.get ("DateTimeOriginal") +
                "</DateTimeOriginal>");
      }

      if (map.containsKey ("DateTime"))
      {
        out.println ("\t\t<DateTime>" + map.get ("DateTime") + "</DateTime>");
      }

      if (map.containsKey ("FocalLength"))
      {
        out.print ("\t\t<FocalLength>");

        try
        {
          out.print (Float.parseFloat (
                  net.grelf.Util.rational (map.get ("FocalLength"))));
        }
        catch (NumberFormatException ex)
        {
          out.print (net.grelf.Util.rational (map.get ("FocalLength")));
        }
        out.println ("</FocalLength>");
      }

      if (map.containsKey ("FNumber"))
      {
        out.print ("\t\t<FNumber>");

        try
        {
          out.print (Float.parseFloat (
                  net.grelf.Util.rational (map.get ("FNumber"))));
        }
        catch (NumberFormatException ex)
        {
          out.print (net.grelf.Util.rational (map.get ("FNumber")));
        }
        out.println ("</FNumber>");
      }

      if (map.containsKey ("ExposureTime"))
      {
        out.print ("\t\t<ExposureTime>");

        try
        {
          out.print (Float.parseFloat (
                  net.grelf.Util.rational (map.get ("ExposureTime"))));
        }
        catch (NumberFormatException ex)
        {
          out.print (net.grelf.Util.rational (map.get ("ExposureTime")));
        }

        out.println ("</ExposureTime>");
      }

      if (map.containsKey ("ISOSpeedRatings"))
      {
        out.println ("\t\t<ISOSpeedRatings>" + map.get ("ISOSpeedRatings") +
                "</ISOSpeedRatings>");
      }
    }

		if (3 == image.getNBands ())
		{
			out.println ("\t\t<Histogram>");
			Histogram hist = new Histogram (image);
			net.grelf.Statistics [] stats = hist.getStatistics ();
			out.println ("\t\t\t<Red>");
			out.println ("\t\t\t\t<Min>" + stats [0].min + "</Min>");
			out.println ("\t\t\t\t<Mode>" + stats [0].mode + "</Mode>");
			out.println ("\t\t\t\t<Max>" + stats [0].max + "</Max>");
			out.println ("\t\t\t\t<Mean>" + stats [0].mean + "</Mean>");
			out.println ("\t\t\t\t<StdDev>" + stats [0].stdDev + "</StdDev>");
			out.println ("\t\t\t</Red>");
			out.println ("\t\t\t<Green>");
			out.println ("\t\t\t\t<Min>" + stats [1].min + "</Min>");
			out.println ("\t\t\t\t<Mode>" + stats [1].mode + "</Mode>");
			out.println ("\t\t\t\t<Max>" + stats [1].max + "</Max>");
			out.println ("\t\t\t\t<Mean>" + stats [1].mean + "</Mean>");
			out.println ("\t\t\t\t<StdDev>" + stats [1].stdDev + "</StdDev>");
			out.println ("\t\t\t</Green>");
			out.println ("\t\t\t<Blue>");
			out.println ("\t\t\t\t<Min>" + stats [2].min + "</Min>");
			out.println ("\t\t\t\t<Mode>" + stats [2].mode + "</Mode>");
			out.println ("\t\t\t\t<Max>" + stats [2].max + "</Max>");
			out.println ("\t\t\t\t<Mean>" + stats [2].mean + "</Mean>");
			out.println ("\t\t\t\t<StdDev>" + stats [2].stdDev + "</StdDev>");
			out.println ("\t\t\t</Blue>");
			out.println ("\t\t</Histogram>");
		}

		// In case the image metadata were incomplete, include any EXIF data found:
		net.grelf.image.Exif.processFile (f.getPath (), out);

		out.println ("\t</Image>");
	} // infoXML

	private javax.swing.ProgressMonitor monitor = null;

	/** For particularly long operations, run them on a worker thread rather than
    * clog up swing's event dispatch thread.
    * Displays a javax.swing.ProgressMonitor which uses file number as a measure
    * of progress. */
	public void doOpInBackground (java.io.File [] files,
                                String fileNameModifier,
	                              java.util.Map <String, Object> params,
                                int nSteps)
	{
		monitor = new javax.swing.ProgressMonitor (null,
                            "Batch [" + this.action + "]", "Running", 0, nSteps);
		monitor.setMillisToDecideToPopup (100);
		OpThread ot = new OpThread (files, this.action, fileNameModifier, params);
		ot.start ();
	} // doOpInBackground

	private class OpThread extends Thread
	{
		private final java.io.File [] files;
		private final BatchAction action;
		private final String fileNameModifier;
		private final java.util.Map <String, Object> params;

		public OpThread (java.io.File [] someFiles, BatchAction anAction,
                     String aFileNameModifier,
		                 java.util.Map <String, Object> someParams)
		{
			files = someFiles;
			action = anAction;
			fileNameModifier = aFileNameModifier;
			params = someParams;
		} // OpThread

		@Override
		public void run ()
		{
      new BatchProcessor (action).batch (files, fileNameModifier, params);
		} // run

	} // OpThread

/*
  public void doAstroPass2LoopInBackground (java.util.List <String> pathsToTempFiles,
          java.util.List<MatchPair> matches, BlobMeas [][] lookup, int nBrightest,
          BatchMetadata batchMetadata)
	{
		monitor = new javax.swing.ProgressMonitor (null, "Batch [astro]", "Running",
                                               0, nFiles);
		monitor.setProgress (0);
		monitor.setMillisToDecideToPopup (100);
		new OpThreadForPass2 (pathsToTempFiles, matches, lookup, nBrightest,
                          batchMetadata).start ();
	} // doAstroPass2LoopInBackground

	private class OpThreadForPass2 extends Thread
	{
    private java.util.List <String> pathsToTempFiles;
    private java.util.List <MatchPair> matches;
    private BlobMeas [][] lookup;
    private int nBrightest;
    private BatchMetadata batchMetadata;

    public OpThreadForPass2 (java.util.List <String> pathsToTempFiles,
            java.util.List <MatchPair> matches, BlobMeas [][] lookup,
            int nBrightest, BatchMetadata batchMetadata)
    {
      this.pathsToTempFiles = pathsToTempFiles;
      this.matches = matches;
      this.lookup = lookup;
      this.nBrightest = nBrightest;
      this.batchMetadata = batchMetadata;
    }

		@Override
		public void run ()
		{
      try
      {
        astroPass2Loop (this.pathsToTempFiles, this.matches, this.lookup,
                        this.nBrightest, this.batchMetadata);
      }
      catch (IncompatibleImageException ex)
      {
        net.grelf.Util.warning ("Error", ex);
      }
		} // run

	} // OpThreadForPass2
*/

	/** Display a new ImFrame in which is drawn the blobs detected in all the
    * images during pass 1 of the astro process. */
	private void displayAllBlobs (int wd, int ht, BlobMeas [][] blobMeases)
	{
    net.grelf.image.Image image =
            net.grelf.image.ImageBase.createImage (wd, ht, 1, 8, true, false);
		int nImages = blobMeases.length;

		for (int imNo = 0; imNo < nImages; imNo++)
		{
			int nBlobs = blobMeases [imNo].length;
			int level = (int) (63.0 + 192.0 * imNo / nImages);
			int [] colour = new int [] { level };

			for (int blobNo = 0; blobNo < nBlobs; blobNo++)
			{
				Blob blob = blobMeases [imNo][blobNo].blob;
				blob.drawRegion (image, colour);
			}
		}

		new ImFrame ("Objects detected in all images", image);
	} // displayAllBlobs

	private void displayBlobsInImage (String filename, BlobMeasList meases)
	{
		ImFrame imf = new ImFrame (meases.size () + " objects detected in " +
            filename, DrawingMode.UNDEFINED,
		        null, -1, wd, ht, new DefaultMeasurementHandler ());

		meases.drawBoundaries (imf);
	} // displayBlobsInImage

} // BatchProcessor