package net.grelf.grip;

// This Java source file is part of the application GRIP, for processing and
// measuring digital photographs.
// GRIP is available free from www.grelf.net - please always quote this URL in
// connection with GRIP.
//
// Copyright (c) Graham T Relf, UK, 2006
//
// All software contains bugs, even if, we hope, they are only extremely subtle
// ones. The author has tested this software to the best of his abilities to
// ensure that it is generally fit for its purpose of processing and measuring
// digital photographs. The author makes absolutely no claims and gives no
// guarantees whatsoever as to the accuracy of any output from this software.
// Anyone making decisions based on any output from this software has a
// responsibility to first convince themselves of the appropriateness of so
// doing. They should test the software for themselves and understand how it
// works, sufficiently for their needs. Source code and explanatory
// documentation are available for that purpose.

/** The menu displayed with a table of measurememts, comparing results from
  * different images. */
public class MeasurementTableMenu
implements java.awt.event.ActionListener, javax.swing.event.MenuListener
{
	private javax.swing.JFrame frame;
	private String [][] data;
	private MeasurementTable table;

	private javax.swing.JMenuBar menubar;
	public javax.swing.JMenuBar getMenuBar () { return menubar; }

	public MeasurementTableMenu (javax.swing.JFrame frame,
                               String [][] data, MeasurementTable table)
	{
    this.frame = frame;
    this.data = data;
    this.table = table;
    initialise ();
  } // MeasurementTableMenu

  private void initialise ()
  {
    javax.swing.JMenu fileMenu = new javax.swing.JMenu ("File");
    fileMenu.setMnemonic ('F');
    fileMenu.add (net.grelf.Util.menuItem (this,
            "Save table as CSV", "savecsv", 'S', 0));
    fileMenu.add (net.grelf.Util.menuItem (this,
            "Save table as XML", "savexml", 'X', java.awt.event.KeyEvent.VK_S));
    fileMenu.add (net.grelf.Util.menuItem (this,
            "Print", "print", 'P', java.awt.event.KeyEvent.VK_P));
    fileMenu.add (net.grelf.Util.menuItem (this,
            "Close window", "close", 'W', java.awt.event.KeyEvent.VK_W));

    javax.swing.JMenu graphMenu = new javax.swing.JMenu ("Graph");
    graphMenu.setMnemonic ('G');
		graphMenu.add (net.grelf.Util.menuItem (this,
            "Graph selected rows", "graph", 'G', 0));

    menubar = new javax.swing.JMenuBar ();
    menubar.add (fileMenu);
    menubar.add (graphMenu);
    menubar.add (new HelpMenu ());

    frame.setJMenuBar (menubar);
  } // initialise

	@Override
	public void actionPerformed (java.awt.event.ActionEvent ev)
	{
		String action = ev.getActionCommand ();

    switch (action)
    {
      case "close":
        frame.setVisible (false);
        frame.dispose ();
        break;

      case "graph":
        MeasurementTableModel model = table.getTableModel ();

        if (null != model)
        {
          int nRows = model.getRowCount ();
          int nCols = model.getColumnCount ();
          java.util.List<MeasurementCurve> curves = new java.util.ArrayList<> ();
          java.util.List<String> imageNames = new java.util.ArrayList<> ();

          for (int colNo = 3; colNo < nCols; colNo++)
          {
            imageNames.add ((String) model.getValueAt (2, colNo));
          }

          for (int rowNo = 3; rowNo < nRows; rowNo++)
          {
            if (table.getJTable ().getSelectionModel ().isSelectedIndex (rowNo))
            {
              String label = (String) model.getValueAt (rowNo, 1);
              String units = (String) model.getValueAt (rowNo, 2);

              if (units.length () > 0)
              {
                label += ", " + units;
              }

              double [] values = new double [nCols - 3];
              boolean ok = true;

              for (int i = 0; i < values.length; i++)
              {
                try
                {
                  values [i] = Double.parseDouble (
                          (String) model.getValueAt (rowNo, i + 3));
                }
                catch (NumberFormatException ex)
                {
                  values [i] = Double.NaN;
                  ok = false;
                }
              }

              if (ok)
              {
                curves.add (new MeasurementCurve (values, label));
              }
              else
              {
                net.grelf.Util.warning ("Warning", label +
                              "<br>has some non-floating-point values<br>so will not be graphed");
              }
            }
          }

          if (curves.size () > 0)
          {
            MeasurementGraph graph = new MeasurementGraph (curves, imageNames);
            javax.swing.JOptionPane.showMessageDialog (
                    frame, graph, "Measured",
                    javax.swing.JOptionPane.PLAIN_MESSAGE,
                    GRIP.getIcon ());
          }
        }

        break;

      case "print":
        java.awt.print.PrinterJob job =
                java.awt.print.PrinterJob.getPrinterJob ();
        job.setPrintable (new ComparisonPrint (data));

        if (job.printDialog (DefaultMeasurementHandler.getAtts ()))
        {
          try
          {
            job.print ();
          }
          catch (java.awt.print.PrinterException ex)
          {
            net.grelf.Util.warning ("Error", ex.toString ());
          }
        }

        break;

      case "savecsv":
        {
          String path = net.grelf.FileIO.selectOutputFile (
                  "Save as CSV file", null,
                   new net.grelf.FileIO_CSVFilter [] {
                            new net.grelf.FileIO_CSVFilter (
                                net.grelf.FileIO_CSVFilter.CSV) },
                   "x.csv");

          if (null != path)
          {
            try
            {
              java.io.PrintWriter out =
                      new java.io.PrintWriter (
                              new java.io.FileWriter (
                                       new java.io.File (path)));

              for (int row = 0; row < data.length; row++)
              {
                for (int col = 0; col < data [0].length; col++)
                {
                  StringBuilder cell = new StringBuilder (data [row][col]);
                  int idx1, idx2;

                  do
                  {
                    idx1 = cell.indexOf ("<");

                    if (idx1 != -1)
                    {
                      idx2 = cell.indexOf (">", idx1);

                      if (idx2 != -1);
                      {
                        cell.delete (idx1, idx2 + 1);
                      }
                    }
                  }
                  while (idx1 != -1);

                  if (-1 != cell.indexOf (","))
                  {
                    out.print ("\"" + cell.toString () + "\",");
                  }
                  else
                  {
                    out.print (cell.toString () + ",");
                  }
                }

                out.println ();
              }

              out.close ();
            }
            catch (java.io.IOException ex)
            {
              net.grelf.Util.message (frame, "Sorry",
                      "Could not open " + path);
            }
          }
          break;
        }
      case "savexml":
        {
          String path = net.grelf.FileIO.selectOutputFile (
                  "Save as XML file", null,
                  new net.grelf.FileIO_XMLFilter [] {
                    new net.grelf.FileIO_XMLFilter (
                            net.grelf.FileIO_XMLFilter.XML) },
                  "x.xml");

          if (null != path)
          {
            MeasurementTableModel mod = table.getTableModel ();

            try
            {
              java.io.PrintWriter out =
                      new java.io.PrintWriter (
                             new java.io.FileWriter (
                                     new java.io.File (path)));
              out.println ("<?xml version=\"1.0\"?>");
              out.println ("<Rows>");

              for (int row = 0; row < data.length; row++)
              {
                out.println ("\t<Row>");

                for (int col = 0; col < data [0].length; col++)
                {
                  String value = data [row][col].trim ();

                  if (0 < value.length ())
                  {
                    String colName = mod.getColumnName (col);
                    out.print ("\t\t<");
                    out.print (colName);
                    out.print (">");
                    out.print (data [row][col]);
                    out.print ("</");
                    out.print (colName);
                    out.println (">");
                  }
                }

                out.println ("\t</Row>");
              }

              out.println ("</Rows>");
              out.close ();
            }
            catch (java.io.IOException ex)
            {
              net.grelf.Util.message (frame, "Sorry", "Could not open " + path);
            }
          }
          break;
        }
    }
	} // actionPerformed

	/** Implementing MenuListener. */
	@Override
	public void menuSelected (javax.swing.event.MenuEvent ev)
	{
	} // menuSelected

	/** Implementing MenuListener but does nothing. */
	@Override
	public void menuDeselected (javax.swing.event.MenuEvent ev) {}

	/** Implementing MenuListener but does nothing. */
	@Override
	public void menuCanceled (javax.swing.event.MenuEvent ev) {}

} // MeasurementTableMenu